#include "opencv2/highgui/highgui.hpp"
#include "opencv2/imgproc/imgproc.hpp"
#include <iostream>
#include <iomanip>
#include <string>
#include <sstream>
#include <stdio.h>
#include <stdlib.h>
#include <algorithm>

using namespace cv;
using namespace std;

Mat src; Mat src_gray;
int thresh = 124;
int max_thresh = 255;
int debug = 0;
RNG rng(12345);
vector<vector<Rect> > lines;

/// Function header
void thresh_callback(int, void* );
bool xComp(Rect shape1, Rect shape2);
bool yVecComp(vector<Rect> shape1, vector<Rect> shape2);
vector<vector<Rect> > sortLines(vector<Rect> total, int height);
vector<Rect> combineExtras(vector<Rect> line);
vector<Rect> gutCharacters(vector<Rect> line, vector<int> holes);
vector<int> findHoles(vector<Rect> line);
bool inRange(Rect theRect, int y);
bool inside(Rect inner, Rect outer);
bool allZero(vector<int> scan);
int averageHeight(vector<Rect> line);
int averageMidpoint(vector<Rect> line);
int getMidPoint(Rect square);

/** @function main */
int main( int argc, char** argv )
{
  /// Load source image and convert it to gray
  src = imread(argv[1], 1);
  int write = atoi(argv[2]);
	if(argc > 3)
		debug = atoi(argv[3]);

  /// Convert image to gray and blur it
  cvtColor( src, src_gray, CV_BGR2GRAY );
  blur( src_gray, src_gray, Size(3,3) );

  /// Create Window
  char source_window[7] = "Source";
  namedWindow( source_window, CV_WINDOW_AUTOSIZE );
  imshow( source_window, src );

  createTrackbar( " Threshold:", "Source", &thresh, max_thresh, thresh_callback );
  thresh_callback( 0, 0 );

  waitKey(0);

	if(write){
		for(int i=0; i < lines.size(); i++){
			for(int j = 0; j < lines[i].size(); j++){
				ostringstream strs;
				string name;

				Mat seg;
				Range ranges[2];
				ranges[0] = Range(lines[i][j].tl().y, lines[i][j].br().y);
				ranges[1] = Range(lines[i][j].tl().x, lines[i][j].br().x);
				seg = src(ranges).clone();
				strs << "letterTests/" << i << "-" << j << ".png";
				name = strs.str();
				cout << "writing: " << name << endl;
				imwrite(name,seg);
			}
		}
	}

  return(0);
}

/** @function thresh_callback */
void thresh_callback(int, void* )
{
  Mat threshold_output;
  vector<vector<Point> > contours;
  vector<Vec4i> hierarchy;
	vector<vector<int> >holes;

  /// Detect edges using Threshold
  threshold(src_gray, threshold_output, thresh, 255, THRESH_BINARY );
  /// Find contours
  findContours( threshold_output, contours, hierarchy, CV_RETR_LIST, CV_CHAIN_APPROX_SIMPLE, Point(0, 0) );

  /// Approximate contours to polygons + get bounding rects and circles
  vector<vector<Point> > contours_poly( contours.size() );
  vector<Rect> boundRect( contours.size() );

  for( int i = 0; i < contours.size(); i++ )
     { approxPolyDP( Mat(contours[i]), contours_poly[i], 3, true );
       boundRect[i] = boundingRect( Mat(contours_poly[i]) );
     }

	int size = contours.size();
  /// Draw polygonal contour + bonding rects + circles
  Mat drawing = Mat::zeros( threshold_output.size(), CV_8UC3 );

  //sort the boundRect
  sort(boundRect.begin(), boundRect.end(), xComp);
	//sort into lines
	lines = sortLines(boundRect, src.rows);

	for(int i=0; i<lines.size(); i++){
		vector<Rect> tmp = combineExtras(lines[i]);
		lines[i] = tmp;
	}

	for(int i=0; i<lines.size(); i++){
		holes.push_back(findHoles(lines[i]));
	}

	int colors[6] = {255, 200, 160, 120, 80, 60};

  // Show in a window
  namedWindow( "Contours", CV_WINDOW_AUTOSIZE );
  if(debug)cout << "Lines: " << lines.size() << endl;
  for( int i = 0; i< lines.size(); i++ ){
  	if(debug)cout << lines[i].size() << endl;
    for(int j = 0; j < lines[i].size(); j++){
    	Scalar color;
    	if(holes[i][j])
    		continue;
    	else
				color = Scalar(colors[i%6], colors[i-1%6], colors[i-2%6] );
			//drawContours( drawing, contours_poly, i, color, 1, 8, vector<Vec4i>(), 0, Point() );
			rectangle( drawing, lines[i][j].tl(), lines[i][j].br(), color, 2, 8, 0 );
			if(debug==2){
				waitKey(0);
				cout 	<< lines[i][j].tl() << " " << lines[i][j].br()<<endl;
				imshow( "Contours", drawing );}
    }
    if(debug)imshow( "Contours", drawing );
    if(debug)waitKey(0);
  }
  imshow( "Contours", drawing );
}

bool xComp(Rect shape1, Rect shape2){
	return ((shape1.tl().x) < (shape2.tl().x));
}

bool yVecComp(vector<Rect> shape1, vector<Rect> shape2){
	return ((shape1[0].tl().y) < (shape2[0].tl().y));
}

vector<vector<Rect> > sortLines(vector<Rect> total, int height){
	total.erase(total.begin()); // the first one is the outline of the image

	vector<vector<Rect> > theLines;
	vector<int> lineScan;

	//scan lines
	for(int y=0; y<height; y++){
		int count=0;
		for(int i=0; i < total.size(); i++){
			if(inRange(total[i], y)){
				count++;
			}
		}
		lineScan.push_back(count);
	}

	//check to make sure lines are being populated
	//for(int i=0; i<lineScan.size(); i++)
	//	cout << i << ": " << lineScan[i] << endl;

	bool flag = true;
	int lastLine = 1;
	while(flag){
		int startRange, endRange;
		startRange = endRange = -1;

		for(int i=lastLine; i < lineScan.size()-1 && endRange < 0; i++){
			if(lineScan[i]==0 && lineScan[i+1]!=0)
				startRange=i;
			else if(lineScan[i]==0 && lineScan[i-1]!=0 && startRange >0){
				endRange=i;
				lastLine = i;
			}
			if(i == lineScan.size()-2 && startRange==-1 && endRange==-1){
				flag = false;
			}
		}
		if(debug)cout << startRange << " " << endRange << endl;

		vector<Rect> aLine;

		if(startRange > 0 && endRange > 0){
			for(int i=0; i < total.size(); i++){
				if(startRange <= total[i].tl().y && endRange >= total[i].br().y){
					aLine.push_back(total[i]);
				}
			}
			sort(aLine.begin(), aLine.end(), xComp);
			theLines.push_back(aLine);

			if(debug)cout <<"Last: "<< lastLine << endl;
		}
	}
	return theLines;
}

vector<Rect> combineExtras(vector<Rect> line){
	vector<Rect>aLine;
	bool skip = false;
	Rect newR;
	for(int i=0; i<line.size(); i++){
		if(skip){
			skip = false;
		}
		else if(i == line.size()-1){
			aLine.push_back(line[i]);
			break;
		}
		else if(line[i].tl().y > line[i+1].br().y){
			if(debug)cout << "new: " << (line[i] | line[i+1]).tl() <<" "<< (line[i] | line[i+1]).br() << endl;
			aLine.push_back(line[i] | line[i+1]);
			skip=true;
		}
		else if(line[i].br().y < line[i+1].tl().y){
			if(debug)cout << "new: " << (line[i] | line[i+1]).tl() <<" "<< (line[i] | line[i+1]).br() << endl;
			aLine.push_back(line[i] | line[i+1]);
			skip=true;
		}
		else
			aLine.push_back(line[i]);
	}
	return aLine;
}

vector<Rect> gutCharacters(vector<Rect> line, vector<int> holes){

}

vector<int> findHoles(vector<Rect> line){
	vector<int> holes;
	holes.push_back(0);
	for(int i=1; i<line.size(); i++){
		if(inside(line[i], line[i-1]) || inside(line[i], line[i-2]))
			holes.push_back(1);
		else
			holes.push_back(0);
	}
	return holes;
}

bool inside(Rect inner, Rect outer){
	if((inner.tl().y > outer.tl().y && inner.tl().x > outer.tl().x) &&
		 (inner.br().y < outer.br().y && inner.br().x < outer.br().x))
		return true;
	else
		return false;
}

bool inRange(Rect theRect, int y){
	if(y >= theRect.tl().y && y <= theRect.br().y){
		return true;
		}
	else{
		return false;
	}
}

bool allZero(vector<int> scan){
	int count = 0;
	int size = scan.size();
	for(int i=0; i<size; i++){
		if(scan[i]==0)
			count++;
	}
	if(count == size)
		return true;
	else
		return false;
}

int averageHeight(vector<Rect> line){
	int average;
	int size = line.size();
	for(int i=0; i < size; i++){
		average += line[i].height;
	}
	return average/size;
}

int averageMidpoint(vector<Rect> line){
	int average;
	int size = line.size();
	for(int i=0; i < size; i++){
		average += getMidPoint(line[i]);
	}
	return average/size;
}

int getMidPoint(Rect square){
	return (square.height/2) + square.tl().y;
}

