/* Main File of Schrift
 *
 * Usage: [input image] [output file] [optional: training 1|0 (default 0)]
 *
 *
 *
 * Lisence Info: ...?
 */

//std libs
#include <iostream>
#include <fstream>
#include <stdio.h>
#include <stdlib.h>
#include <vector>
#include <string>

//openCV
#include <cv.h>
#include <ml.h>
#include <highgui.h>

//namespaces
using namespace std;
using namespace cv;

#define TOP 		1
#define BOTTOM 	2
#define LEFT	 	3
#define RIGHT 	4

CvANN_MLP machineBrain;
int debug = 0;
vector<vector<int> > space;
//
// HELPER FUNCTIONS
//
bool inRange(Rect theRect, int y){
	if(y >= theRect.tl().y && y <= theRect.br().y){
		return true;
		}
	else{
		return false;
	}
}

bool xComp(Rect shape1, Rect shape2){
	return ((shape1.tl().x) < (shape2.tl().x));
}

bool inside(Rect inner, Rect outer){
	if((inner.tl().y > outer.tl().y && inner.tl().x > outer.tl().x) &&
		 (inner.br().y < outer.br().y && inner.br().x < outer.br().x))
		return true;
	else
		return false;
}

float blackPixelCount(Mat img){
	int all = img.cols * img.rows;
	int black = all-countNonZero(img);
	return black;
}

Mat getHalf(Mat img, int which){
	Range ranges[2];
	//if(debug)cout << img.rows << " " << img.cols << endl;
	switch(which){
		case 1: //top
			ranges[0] = Range(0, (float)img.rows/2);
			ranges[1] = Range(0, img.cols);
			break;
		case 2: //bottom
			ranges[0] = Range((float)img.rows/2, img.rows);
			ranges[1] = Range(0, img.cols);
			break;
		case 3: //left
			ranges[0] = Range(0, img.rows);
			ranges[1] = Range(0, (float)img.cols/2);
			break;
		case 4: //right
			ranges[0] = Range(0, img.rows);
			ranges[1] = Range((float)img.cols/2, img.cols);
			break;
		default:
			break;
	}
	//if(debug)cout << ranges[0].start <<" "<< ranges[0].end << ", "
	//		<< ranges[1].start <<" "<< ranges[1].end << endl << endl;
	if(ranges[0].end == 0)
		ranges[0].end = 1;
	if(ranges[1].end == 0)
		ranges[1].end = 1;
	return img(ranges).clone();
}

//
// IMAGE AND SEGMENTATION FUNNCTIONS
//
vector<Rect> combineExtras(vector<Rect> line){
	vector<Rect>aLine;
	bool skip = false;
	Rect newR;
	for(int i=0; i<line.size(); i++){
		if(skip){
			skip = false;
		}
		else if(i == line.size()-1){
			aLine.push_back(line[i]);
			break;
		}
		else if(line[i].tl().y > line[i+1].br().y){
			if(debug)cout << "new: " << (line[i] | line[i+1]).tl() <<" "<< (line[i] | line[i+1]).br() << endl;
			aLine.push_back(line[i] | line[i+1]);
			skip=true;
		}
		else if(line[i].br().y < line[i+1].tl().y){
			if(debug)cout << "new: " << (line[i] | line[i+1]).tl() <<" "<< (line[i] | line[i+1]).br() << endl;
			aLine.push_back(line[i] | line[i+1]);
			skip=true;
		}
		else
			aLine.push_back(line[i]);
	}
	return aLine;
}

vector<int> findHoles(vector<Rect> line){
	vector<int> holes;
	holes.push_back(0);
	for(int i=1; i<line.size(); i++){
		if(inside(line[i], line[i-1]) || inside(line[i], line[i-2]))
			holes.push_back(1);
		else
			holes.push_back(0);
	}
	return holes;
}

vector<vector<Rect> > sortLines(vector<Rect> total, int height){
	total.erase(total.begin()); // the first one is the outline of the image

	vector<vector<Rect> > theLines;
	vector<int> lineScan;

	//scan lines
	for(int y=0; y<height; y++){
		int count=0;
		for(int i=0; i < total.size(); i++){
			if(inRange(total[i], y)){
				count++;
			}
		}
		lineScan.push_back(count);
	}

	//check to make sure lines are being populated
	//for(int i=0; i<lineScan.size(); i++)
	//	if(debug)cout << i << ": " << lineScan[i] << endl;

	bool flag = true;
	int lastLine = 1;
	while(flag){
		int startRange, endRange;
		startRange = endRange = -1;

		for(int i=lastLine; i < lineScan.size()-1 && endRange < 0; i++){
			if(lineScan[i]==0 && lineScan[i+1]!=0)
				startRange=i;
			else if(lineScan[i]==0 && lineScan[i-1]!=0 && startRange >0){
				endRange=i;
				lastLine = i;
			}
			if(i == lineScan.size()-2 && startRange==-1 && endRange==-1){
				flag = false;
			}
		}
		 //if(debug)cout << startRange << " " << endRange << endl;

		vector<Rect> aLine;

		if(startRange > 0 && endRange > 0){
			for(int i=0; i < total.size(); i++){
				if(startRange <= total[i].tl().y && endRange >= total[i].br().y){
					aLine.push_back(total[i]);
				}
			}
			sort(aLine.begin(), aLine.end(), xComp);
			theLines.push_back(aLine);

			 //if(debug)cout <<"Last: "<< lastLine << endl;
		}
	}
	return theLines;
}

//lower light images won't work well for now I figure there is a simple
// algorithm to fix this based on mean image color
Mat threshold(Mat image){
	Mat ret;
	int thresh = 124;
	int max 	 = 255;

	cvtColor( image, ret, CV_RGB2GRAY );
	threshold( ret, ret, thresh, max, THRESH_BINARY);

	return ret;
}

//takes the thresholded image and segments characters
//sorts in order them by line.
vector<vector<Mat> > segmentation(Mat image){
	vector<vector<Mat> > charactersByLine;
	vector<vector<Point> > contours;
	vector<Vec4i> hierarchy;
	vector<vector<int> > holes;

  /// Find contours
  findContours( image, contours, hierarchy, CV_RETR_LIST, CV_CHAIN_APPROX_SIMPLE, Point(0, 0) );

  /// Approximate contours to polygons + get bounding rects and circles
	vector<vector<Point> > contours_poly( contours.size() );
  vector<Rect> boundRect( contours.size() );

	//get bounding box
	for( int i = 0; i < contours.size(); i++ ){
		approxPolyDP( Mat(contours[i]), contours_poly[i], 3, true );
		boundRect[i] = boundingRect( Mat(contours_poly[i]) );
	}

	//sort the boundRect
  sort(boundRect.begin(), boundRect.end(), xComp);
	//sort into lines
	vector<vector<Rect> > lines = sortLines(boundRect, image.rows);

	for(int i=0; i<lines.size(); i++){
		vector<Rect> tmp = combineExtras(lines[i]);
		lines[i] = tmp;
	}

	for(int i=0; i<lines.size(); i++){
		holes.push_back(findHoles(lines[i]));
	}

	//for each bounding box, clone that section from image and push it to letters
	for(int i=0; i<lines.size(); i++){
		vector<Mat> letters;
		vector<int> spacetmp;
		for(int j=0; j<lines[i].size();j++){
			if(holes[i][j])
				continue;
			Mat seg;
			Range ranges[2];
			ranges[0] = Range(lines[i][j].tl().y, lines[i][j].br().y);
			ranges[1] = Range(lines[i][j].tl().x, lines[i][j].br().x);
			seg = image(ranges).clone();
			letters.push_back(seg);
			if(j!=lines[i].size()-1 && lines[i][j].br().x+lines[i][j].width < lines[i][j+1].tl().x)
				spacetmp.push_back(1);
			else
				spacetmp.push_back(0);
		}
		space.push_back(spacetmp);
		if(debug)cout << letters.size() << " letters in line " << i+1 << endl;
		charactersByLine.push_back(letters);
	}

	//return vector
	return charactersByLine;
}

//
// NEURAL NETWORK
//
void trainMachine(){

	//The training data vectors.
	vector<Mat> polyLetters;
	vector<char> letters;

	//Read the training file
	//has the format: [file name] [string to produce]
	ifstream trainer;
	trainer.open("trainer.txt");
	string line;

	//Get the number of samples.
	getline(trainer, line);
	istringstream convert(line);
	int trainSamples;
	convert >> trainSamples;
	if(debug)cout << "Training with "<< trainSamples << " line samples...\n";

	int inputSize = 5;
	int outputSize = 1;
	int layers = 5;
	int nodesPerHLayer = 5;

	//Read and populate the samples.
	for (int i=0; !trainer.eof(); i++){
		//get name of line from line and open it.
		getline(trainer, line);
		size_t pos = line.find_first_of(' ');
		string testImageName = line.substr(0, pos);
		if(debug)cout << "reading: " << testImageName << endl;
		Mat tempImage = imread(testImageName, 1);
		Mat image = threshold(tempImage);

		//segment letters.
		vector<vector<Mat> > theLine = segmentation(image);
		if(debug)cout << "\tsegmented" << endl;
		for(int i = 0; i< theLine.size(); i++){
			for(int j=0; j< theLine[i].size(); j++){
				polyLetters.push_back(theLine[i][j]);
			}
		}
		//populate testing array
		if(debug)cout << "\tPushed: ";
		for(int i=pos; i < line.length(); i++){
			if(line[i]!=' ')
				if(debug)cout << line[i];
				letters.push_back(line[i]);
		}
		if(debug)cout << endl;
	}
	trainer.close();

	trainSamples = letters.size();
	//Create training matrices
	//Input data samples.
	CvMat* trainData = cvCreateMat(trainSamples, inputSize, CV_32FC1);

	//Output data samples. Matrix of order (train_sample_count x 1)
	CvMat* trainClasses = cvCreateMat(trainSamples, outputSize, CV_32FC1);

	//The weight of each training data sample. We'll later set all to equal weights.
	CvMat* sampleWts = cvCreateMat(trainSamples, 1, CV_32FC1);

	CvMat* neuralLayers = cvCreateMat(layers, 1, CV_32SC1);

	CvMat trainData1, trainClasses1, neuralLayers1, sampleWts1;

	cvGetRows(trainData, &trainData1, 0, trainSamples);
	cvGetRows(trainClasses, &trainClasses1, 0, trainSamples);
	cvGetRows(trainClasses, &trainClasses1, 0, trainSamples);
	cvGetRows(sampleWts, &sampleWts1, 0, trainSamples);
	cvGetRows(neuralLayers, &neuralLayers1, 0, layers);

	//Setting the number of neurons on each layer of the ANN
	cvSet1D(&neuralLayers1, 0, cvScalar(inputSize));

	for(int i=1; i < layers-1; i++)
		cvSet1D(&neuralLayers1, i, cvScalar(nodesPerHLayer));

	cvSet1D(&neuralLayers1, layers-1, cvScalar(outputSize));

	//Assemble the ML training data.
	for(int i=0; i< trainSamples; i++)
	{
		//training input
		//training input
		cvSetReal2D(&trainData1, i, 0, blackPixelCount(getHalf(polyLetters[i], TOP)));
		cvSetReal2D(&trainData1, i, 1, blackPixelCount(getHalf(polyLetters[i], BOTTOM)));
		cvSetReal2D(&trainData1, i, 2, blackPixelCount(getHalf(polyLetters[i], LEFT)));
		cvSetReal2D(&trainData1, i, 3, blackPixelCount(getHalf(polyLetters[i], RIGHT)));
		cvSetReal2D(&trainData1, i, 4, polyLetters[i].rows/polyLetters[i].cols);

		//Output for char
		cvSet1D(&trainClasses1, i, cvScalar((int)letters[i]));
		//Weight (setting everything to 1)
		cvSet1D(&sampleWts1, i, cvScalar(1));
	}
	if(debug)cout << "Training arrays filled" << endl;

	//Create our ANN.
	machineBrain.create(neuralLayers);

	if(debug)cout << "Training ..." << endl;
	//Train it with our data.
	machineBrain.train(
		trainData,
		trainClasses,
		sampleWts,
		0,
		CvANN_MLP_TrainParams(
				cvTermCriteria(
						CV_TERMCRIT_ITER+CV_TERMCRIT_EPS,
						100000,
						0.001 // error rate
						),
				CvANN_MLP_TrainParams::BACKPROP,
				0.01,
				0.05
				)
		);
		if(debug)cout << "Trained." << endl << endl;
}

void predict(vector<vector <Mat> > lines){
	cout << endl;
	for(int i=0; i < lines.size(); i++){
		for(int j=0; j< lines[i].size(); j++){
				float _sample[5];
				CvMat sample = cvMat(1, 5, CV_32FC1, _sample);
				sample.data.fl[0] = blackPixelCount(getHalf(lines[i][j], TOP));
				sample.data.fl[1] = blackPixelCount(getHalf(lines[i][j], BOTTOM));
				sample.data.fl[2] = blackPixelCount(getHalf(lines[i][j], LEFT));
				sample.data.fl[3] = blackPixelCount(getHalf(lines[i][j], RIGHT));
				sample.data.fl[4] = lines[i][j].rows/lines[i][j].cols;

				float _predout[1];
				CvMat predout = cvMat(1, 1, CV_32FC1, _predout);

				//if(debug)cout << "predicting..." << endl;
				machineBrain.predict(&sample, &predout);
				cout << (char)predout.data.fl[0];

				if(space[i][j])
					cout << " ";
		}
		cout << endl << endl;
	}
}


/*
 *_________MAIN_________
 */
int main(int argc, char * argv[]){

  if(argc < 1){
    cout << "Usage: [input image] [optional: debug]\n";
    exit(1);
  }

	if(argc > 2)
		debug = atoi(argv[2]);

  char* imageName  = argv[1];

  Mat inputImage;
  Mat bitImage;
  inputImage = imread( imageName, 1 );

	bitImage = threshold(inputImage);

  vector<vector<Mat> > characters = segmentation(bitImage);
	if(debug) cout << "Lines: " << characters.size() << endl;

	trainMachine();
	predict(characters);

  return 0;
}
